import logging
import queue
import threading
from typing import Optional

import requests

from cloner.put_repos_in_queue import put_repos_in_queue

logger = logging.getLogger(__file__)


class TokenNotFoundForGHEException(Exception):
    def __init__(self, message="A token is needed when cloning from a GHE server"):
        super().__init__(message)


def obtain_repos(
    github_organization: str,
    github_token: Optional[str],
    queue_lock: threading.Lock,
    repo_queue: queue.Queue,
    ghe: str = None,
) -> None:
    """
    Makes a request to the GitHub API and then passes that info to the queue.
    GitHub docs: https://docs.github.com/en/rest?apiVersion=2022-11-28

    Docs for GitHub Enterprise available under: https://docs.github.com/en/enterprise-server@3.7
    Or any matching version that applies
    """

    if ghe:
        if github_token is None:
            raise TokenNotFoundForGHEException
        github_url = f"https://{ghe}/api/v3/orgs/{github_organization}/repos"
        headers = {
            "Accept": "application/vnd.github+json",
            "Authorization": f"Bearer {github_token}",
        }
    else:
        github_url = f"https://api.github.com/orgs/{github_organization}/repos"
        headers = {
            "Accept": "application/vnd.github.v3+json",
            "X-GitHub-Api-Version": "2022-11-28",
        }
        if github_token:
            headers["Authorization"] = f"token {github_token}"

    params = {"per_page": 100}

    logger.debug("First call to GitHub")
    response = requests.get(github_url, headers=headers, params=params)
    response.raise_for_status()
    json_response = response.json()

    while "next" in response.links.keys():
        logger.debug("Keep calling GitHub")
        response = requests.get(response.links["next"]["url"], headers=headers, params=params)
        json_response.extend(response.json())

    logger.debug("Finished calling GitHub")
    put_repos_in_queue(json_response, queue_lock, repo_queue)



import os
from multiprocessing import Process
from typing import Optional

from cloner.repository import Repository


class ClonerProcess(Process):
    """Class to clone a list of given repositories when started."""

    def __init__(
        self,
        repos_to_clone: list[Repository],
        process_id: int,
        clone_path: Optional[str] = None,
        git_options: Optional[str] = None,
    ):
        Process.__init__(self)
        self.repos_list = repos_to_clone
        self.process_id = process_id
        self._clone_path = clone_path
        self._git_options = git_options

    @property
    def clone_path(self):
        if self._clone_path is not None:
            return self._clone_path
        return "./repos/"

    @property
    def git_options(self):
        default_option = "--quiet "
        if self._git_options is not None:
            return default_option + self._git_options + " "
        return default_option

    def _execute_system_command(self, command):
        return os.system(command=command)

    def run(self):
        """Clones each repo given in the constructor."""
        for repo in self.repos_list:
            self._execute_system_command(
                command=f"git clone {self.git_options}{repo.clone_url} {self.clone_path}{repo.name}"
            )



import logging
import queue
import threading

import click
from requests import HTTPError

from cloner.__version__ import __version__
from cloner.clone_repos import clone_repos
from cloner.obtain_repos import obtain_repos
from cloner.split_queue import split_queue

repository_list_queue_lock = threading.Lock()
repository_list_queue = queue.Queue()

LOGGING_LEVELS = {
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}


def setup_logging(level: str) -> None:
    """Logging setup and configuration."""
    logging.basicConfig(level=LOGGING_LEVELS[level], format="%(levelname)s - %(message)s")


@click.command()
@click.version_option(prog_name="cloner", version=__version__)
@click.argument("github_organization")
@click.option(
    "--token",
    "token",
    type=str,
    default=None,
    help="GitHub token to read private repos. This parameter is needed when cloning from an GitHub Enterprise server.",
    show_default=True,
)
@click.option(
    "--ghe",
    "github_enterprise",
    type=str,
    default=None,
    help="GitHub Enterprise URL. "
    "It needs the GITHUB_ORGANIZATION parameter to clone repos from there and the TOKEN option as well.",
    show_default=True,
)
@click.option(
    "--threads",
    "threads",
    type=int,
    default=4,
    help="Number of threads and processes to use.",
    show_default=True,
)
@click.option(
    "--logging",
    "logging_level",
    type=click.Choice(LOGGING_LEVELS.keys(), case_sensitive=True),
    default="INFO",
    help="Logging level",
    show_default=True,
)
@click.option(
    "--path",
    "clone_path",
    type=str,
    default=None,
    help="Sets a path where to clone the repositories (eg: ./another/path/)",
    show_default=True,
)
@click.option(
    "--git-options",
    "git_options",
    type=str,
    default=None,
    help='Add options to the clone command (eg: --git-options "--depth 1"). By default, clones quietly (--quiet).',
    show_default=True,
)
def cli(
    github_organization: str,
    token: str,
    github_enterprise: str,
    threads: int,
    logging_level: str,
    clone_path: str,
    git_options: str,
) -> None:
    """Clones all visible repositories for a given organization."""
    setup_logging(level=logging_level)

    logging.info(f"Cloning repos for: {github_organization}")

    try:
        obtain_repos(
            github_organization=github_organization,
            github_token=token,
            queue_lock=repository_list_queue_lock,
            repo_queue=repository_list_queue,
            ghe=github_enterprise,
        )
    except HTTPError as e:
        logging.error("An error has occurred while obtaining repos", exc_info=e)

    total_repos_to_clone = repository_list_queue.qsize()
    logging.info(f"Total repos to clone: {total_repos_to_clone}")

    if total_repos_to_clone > 0:
        repos_to_clone = split_queue(
            number_of_threads=threads,
            repository_queue=repository_list_queue,
            repository_queue_lock=repository_list_queue_lock,
        )

        logging.info("Cloning repos...")

        clone_repos(
            number_of_threads=threads,
            repos_to_clone=repos_to_clone,
            clone_path=clone_path,
            git_options=git_options,
        )

        logging.info("Repos cloned!")
